# -*- coding: utf-8 -*-
# Copyright (C) 2017-2019, 2021 Arno Onken
#
# This file is part of the mixedvines package.
#
# The mixedvines package is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# The mixedvines package is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
'''
This module implements a copula vine model with mixed marginals.
'''
from scipy.stats import kendalltau, norm, uniform
from scipy.optimize import minimize
import numpy as np
from .marginal import Marginal
from .copula import Copula, IndependenceCopula


class MixedVine(object):
    '''
    This class represents a copula vine model with mixed marginals.

    Parameters
    ----------
    dim : int
        The number of marginals of the vine model.  Must be greater than 1.

    Attributes
    ----------
    root : VineLayer
        The root layer of the vine tree.

    Methods
    -------
    logpdf(samples)
        Calculates the log of the probability density function.
    pdf(samples)
        Calculates the probability density function.
    rvs(size, random_state)
        Generates random variates from the mixed vine.
    entropy(alpha, sem_tol, mc_size, random_state)
        Estimates the entropy of the mixed vine.
    set_marginal(marginal_index, rv_mixed)
        Sets a particular marginal distribution in the mixed vine tree.
    set_copula(layer_index, copula_index, copula)
        Sets a particular pair copula in the mixed vine tree.
    is_continuous()
        Determines which marginals are continuous.
    fit(samples, is_continuous, trunc_level, do_refine, keep_order)
        Fits the mixed vine to the given samples.
    '''

    class VineLayer(object):
        '''
        This class represents a layer of a copula vine tree.  A tree
        description in layers is advantageous, because most operations on the
        vine work in sweeps from layer to layer.

        Parameters
        ----------
        input_layer : VineLayer, optional
            The layer providing input.  (Default: `None`)
        input_indices : array_like, optional
            Array of length n where n is the number of copulas in this layer.
            Each element in the array is a 2-tuple containing the left and
            right input indices of the respective pair-copula.  `None` if this
            is the marginal layer.
        marginals : array_like, optional
            List with the marginal distributions as elements.  `None` if this
            is not the marginal layer.
        copulas : array_like, optional
            List with the pair-copulas of this layer as elements.  `None` if
            this is the marginal layer.

        Attributes
        ----------
        input_layer : VineLayer
            The layer providing input.
        output_layer : VineLayer
            The output layer.
        input_indices : array_like
            Array with the input indices of the copulas.
        input_marginal_indices : array_like
            Array with the input indices of the marginals.
        marginals : array_like
            List with the marginal distributions of this layer as elements.
        copulas : array_like
            List with the pair-copulas of this layer as elements.

        Methods
        -------
        is_marginal_layer()
            Determines whether the layer is a marginal layer.
        is_root_layer()
            Determines whether the layer is a root layer.
        logpdf(samples)
            Log of the probability density function.
        densities(samples)
            Computes densities and cumulative distribution functions.
        build_curvs(urvs, curvs)
            Builds conditional uniform random variates `curvs` for
            `make_dependent`.
        curv_ccdf(sample, curvs, copula_index)
            Generates a conditional sample for `build_curvs`.
        make_dependent(urvs, curvs)
            Introduces dependencies between the uniform random variates `urvs`.
        rvs(size, random_state)
            Generates random variates from the mixed vine.
        fit(samples, is_continuous, trunc_level)
            Fits a vine tree.
        get_all_params()
            Constructs an array containing all copula parameters.
        set_all_params(params)
            Sets all copula parameters to the values stored in params.
        get_all_bounds()
            Collects the bounds of all copula parameters.
        is_continuous()
            Determines which marginals are continuous.
        '''

        def __init__(self, input_layer=None, input_indices=None,
                     marginals=None, copulas=None):
            self.input_layer = input_layer
            self.output_layer = None
            if input_layer is not None:
                input_layer.output_layer = self
            self.input_indices = input_indices
            self.marginals = marginals
            self.copulas = copulas
            # Set indices of input marginals
            if input_indices is not None:
                if input_layer.is_marginal_layer():
                    self.input_marginal_indices = input_indices
                else:
                    self.input_marginal_indices = []
                    for _, i_ind in enumerate(input_indices):
                        self.input_marginal_indices.append(np.array([
                            input_layer.input_marginal_indices[i_ind[0]][1],
                            input_layer.input_marginal_indices[i_ind[1]][1]]))
            else:
                self.input_marginal_indices = None

        def is_marginal_layer(self):
            '''
            Determines whether the layer is the marginal layer.

            Returns
            -------
            boolean
                `True` if the layer is the marginal layer.
            '''
            return not self.input_layer

        def is_root_layer(self):
            '''
            Determines whether the layer is the output layer.

            Returns
            -------
            boolean
                `True` if the layer is the root layer.
            '''
            return not self.output_layer

        def logpdf(self, samples):
            '''
            Calculates the log of the probability density function.

            Parameters
            ----------
            samples : array_like
                n-by-d matrix of samples where n is the number of samples and d
                is the number of marginals.

            Returns
            -------
            ndarray
                Log of the probability density function evaluated at `samples`.
            '''
            if samples.size == 0:
                return np.empty((0, 1))
            if self.is_root_layer():
                res = self.densities(samples)
                return res['logpdf']
            return self.output_layer.logpdf(samples)

        def _marginal_densities(self, samples):
            '''
            Evaluate marginal densities and cumulative distribution functions.

            Parameters
            ----------
            samples : array_like
                n-by-d matrix of samples where n is the number of samples and d
                is the number of marginals.

            Returns
            -------
            dout : dictionary
                The densities and cumulative distribution functions.  Keys:
                `logpdf`: Equal to first element of `logp`.
                'logp': Log of the probability density function.
                'cdfp': Upper cumulative distribution functions.
                'cdfm': Lower cumulative distribution functions.
                'is_continuous': List of booleans where element i is `True` if
                output element i is continuous.
            '''
            logp = np.zeros(samples.shape)
            cdfp = np.zeros(samples.shape)
            cdfm = np.zeros(samples.shape)
            is_continuous = np.zeros(len(self.marginals), dtype=bool)
            for k, marginal in enumerate(self.marginals):
                is_continuous[k] = marginal.is_continuous
                cdfp[:, k] = marginal.cdf(samples[:, k])
                if marginal.is_continuous:
                    logp[:, k] = marginal.logpdf(samples[:, k])
                else:
                    cdfm[:, k] = marginal.cdf(samples[:, k] - 1)
                    old_settings = np.seterr(divide='ignore')
                    logp[:, k] = np.log(np.maximum(0, cdfp[:, k] - cdfm[:, k]))
                    np.seterr(**old_settings)
            logpdf = logp[:, self.output_layer.input_indices[0][0]]
            dout = {'logpdf': logpdf, 'logp': logp, 'cdfp': cdfp, 'cdfm': cdfm,
                    'is_continuous': is_continuous}
            return dout

        def densities(self, samples):
            '''
            Computes densities and cumulative distribution functions layer by
            layer.

            Parameters
            ----------
            samples : array_like
                n-by-d matrix of samples where n is the number of samples and d
                is the number of marginals.

            Returns
            -------
            dout : dictionary
                The densities and cumulative distribution functions.  Keys:
                `logpdf`: Sum of the first elements of `logp` of all input
                layers and this one.
                'logp': Log of the probability density function.
                'cdfp': Upper cumulative distribution functions.
                'cdfm': Lower cumulative distribution functions.
                'is_continuous': List of booleans where element i is `True` if
                output element i is continuous.
            '''
            if self.is_marginal_layer():
                return self._marginal_densities(samples)
            # Propagate samples to input_layer
            din = self.input_layer.densities(samples)
            # Prepare output densities
            logp = np.zeros((samples.shape[0], len(self.copulas)))
            cdfp = np.zeros((samples.shape[0], len(self.copulas)))
            cdfm = np.zeros((samples.shape[0], len(self.copulas)))
            is_continuous = np.zeros(len(self.copulas), dtype=bool)
            for k, copula in enumerate(self.copulas):
                i = self.input_indices[k][0]
                j = self.input_indices[k][1]
                # Distinguish between discrete and continuous inputs
                if din['is_continuous'][i] and din['is_continuous'][j]:
                    cdfp[:, k] \
                        = copula.ccdf(
                            np.array([din['cdfp'][:, i],
                                      din['cdfp'][:, j]]).T, axis=0)
                    logp[:, k] \
                        = copula.logpdf(
                            np.array([din['cdfp'][:, i],
                                      din['cdfp'][:, j]]).T) \
                        + din['logp'][:, j]
                elif not din['is_continuous'][i] \
                        and din['is_continuous'][j]:
                    old_settings = np.seterr(divide='ignore')
                    isf = np.isfinite(din['logp'][:, i])
                    cdfp[~isf, k] = 0.0
                    cdfp[isf, k] = np.exp(np.log(
                        np.maximum(0,
                                   copula.cdf(
                                       np.array([din['cdfp'][isf, i],
                                                 din['cdfp'][isf, j]]).T)
                                   - copula.cdf(
                                       np.array([din['cdfm'][isf, i],
                                                 din['cdfp'][isf, j]]).T)))
                                          - din['logp'][isf, i])
                    isf = np.isfinite(din['logp'][:, i])
                    logp[~isf, k] = -np.inf
                    logp[isf, k] = np.log(
                        np.maximum(0,
                                   copula.ccdf(
                                       np.array([din['cdfp'][isf, i],
                                                 din['cdfp'][isf, j]]).T)
                                   - copula.ccdf(
                                       np.array([din['cdfm'][isf, i],
                                                 din['cdfp'][isf, j]]).T))
                        ) - din['logp'][isf, i] + din['logp'][isf, j]
                    np.seterr(**old_settings)
                elif din['is_continuous'][i] \
                        and not din['is_continuous'][j]:
                    cdfp[:, k] \
                        = copula.ccdf(
                            np.array([din['cdfp'][:, i],
                                      din['cdfp'][:, j]]).T, axis=0)
                    cdfm[:, k] \
                        = copula.ccdf(
                            np.array([din['cdfp'][:, i],
                                      din['cdfm'][:, j]]).T, axis=0)
                    old_settings = np.seterr(divide='ignore')
                    logp[:, k] \
                        = np.log(np.maximum(0, cdfp[:, k] - cdfm[:, k]))
                    np.seterr(**old_settings)
                else:
                    old_settings = np.seterr(divide='ignore')
                    isf = np.isfinite(din['logp'][:, i])
                    cdfp[~isf, k] = 0.0
                    cdfp[isf, k] = np.exp(np.log(
                        np.maximum(0,
                                   copula.cdf(
                                       np.array([din['cdfp'][isf, i],
                                                 din['cdfp'][isf, j]]).T)
                                   - copula.cdf(
                                       np.array([din['cdfm'][isf, i],
                                                 din['cdfp'][isf, j]]).T)))
                                          - din['logp'][isf, i])
                    isf = np.isfinite(din['logp'][:, i])
                    cdfm[~isf, k] = 0.0
                    cdfm[isf, k] = np.exp(np.log(
                        np.maximum(0,
                                   copula.cdf(
                                       np.array([din['cdfp'][isf, i],
                                                 din['cdfm'][isf, j]]).T)
                                   - copula.cdf(
                                       np.array([din['cdfm'][isf, i],
                                                 din['cdfm'][isf, j]]).T)))
                                          - din['logp'][isf, i])
                    logp[:, k] \
                        = np.log(np.maximum(0, cdfp[:, k] - cdfm[:, k]))
                    np.seterr(**old_settings)
                # This propagation of continuity is specific for the c-vine
                is_continuous[k] = din['is_continuous'][j]
            logpdf = din['logpdf'] + logp[:, 0]
            dout = {'logpdf': logpdf, 'logp': logp, 'cdfp': cdfp, 'cdfm': cdfm,
                    'is_continuous': is_continuous}
            return dout

        def build_curvs(self, urvs, curvs):
            '''
            Helper function for `make_dependent`.  Builds conditional uniform
            random variates `curvs` for `make_dependent`.

            Parameters
            ----------
            urvs : array_like
                Uniform random variates to be made dependent by
                `make_dependent`.
            curvs : array_like
                Array to be filled with dependent conditional uniform random
                variates by `make_dependent`.

            Returns
            -------
            urvs : array_like
                Dependent uniform random variates.
            curvs : array_like
                Conditional uniform random variates.
            '''
            (urvs, curvs) = self.make_dependent(urvs, curvs)
            if self.is_marginal_layer():
                first_marginal_index = self.output_layer.input_indices[0][0]
                curvs[:, first_marginal_index] = urvs[:, first_marginal_index]
            else:
                copula_index = 0
                curv_index = self.input_marginal_indices[copula_index][1]
                curvs[:, curv_index] = self.curv_ccdf(urvs[:, curv_index],
                                                      curvs, copula_index)
            return (urvs, curvs)

        def curv_ccdf(self, sample, curvs, copula_index):
            '''
            Helper function for `build_curvs` to generate a conditional sample.

            Parameters
            ----------
            sample : float
                Right input for the marginal layer.
            curvs : array_like
                Conditional uniform random variates.
            copula_index : int
                Index of the copula to be used to generate the dependent
                sample.

            Returns
            -------
            sample : float
                Conditional sample for `curvs` at index `copula_index`.
            '''
            if not self.is_marginal_layer():
                sample = self.input_layer.curv_ccdf(
                    sample, curvs, self.input_indices[copula_index][1])
                curv_index = self.input_marginal_indices[copula_index][0]
                input_urvs = np.array([curvs[:, curv_index], sample]).T
                sample = self.copulas[copula_index].ccdf(input_urvs, axis=0)
            return sample

        def make_dependent(self, urvs, curvs=None):
            '''
            Helper function for `rvs`.  Introduces dependencies between the
            uniform random variates `urvs` according to the vine copula tree.

            Parameters
            ----------
            urvs : array_like
                Uniform random variates to be made dependent.
            curvs : array_like, optional
                Array to be filled with dependent conditional uniform random
                variates by `build_curvs`.  (Default: `None`)

            Returns
            -------
            urvs : array_like
                Dependent uniform random variates.
            curvs : array_like
                Conditional uniform random variates.
            '''
            if curvs is None:
                curvs = np.zeros(shape=urvs.shape)
            if not self.is_marginal_layer():
                (urvs, curvs) = self.input_layer.build_curvs(urvs, curvs)
            copula_index = 0
            layer = self
            while not layer.is_marginal_layer():
                imi = layer.input_marginal_indices[copula_index]
                input_urvs = np.array([curvs[:, imi[0]], urvs[:, imi[1]]]).T
                urvs[:, imi[1]] = layer.copulas[copula_index].ppcf(input_urvs,
                                                                   axis=0)
                copula_index = layer.input_indices[copula_index][1]
                layer = layer.input_layer
            return (urvs, curvs)

        def rvs(self, size=1, random_state=None):
            '''
            Generates random variates from the mixed vine.  Currently assumes a
            c-vine structure.

            Parameters
            ----------
            size : int, optional
                The number of samples to generate.  (Default: 1)
            random_state : {None, int, RandomState, Generator}, optional
                The random state to use for random variate generation.  `None`
                corresponds to the `RandomState` singleton.  For an int, a
                new `RandomState` is generated and seeded.  For a `RandomState`
                or `Generator`, the object is used.  (Default: `None`)

            Returns
            -------
            array_like
                n-by-d matrix of samples where n is the number of samples and d
                is the number of marginals.
            '''
            if self.is_root_layer():
                # Determine distribution dimension
                layer = self
                while not layer.is_marginal_layer():
                    layer = layer.input_layer
                dim = len(layer.marginals)
                samples = uniform.rvs(size=[size, dim],
                                      random_state=random_state)
                (samples, _) = self.make_dependent(samples)
                # Use marginals to transform dependent uniform samples
                for i, marginal in enumerate(layer.marginals):
                    samples[:, i] = marginal.ppf(samples[:, i])
                return samples
            return self.output_layer.rvs(size=size, random_state=random_state)

        def fit(self, samples, is_continuous, trunc_level=None):
            '''
            Fits the vine tree to the given samples.  This method is supposed
            to be called on the output layer and will recurse to its input
            layers.

            Parameters
            ----------
            samples : array_like
                n-by-d matrix of samples where n is the number of samples and d
                is the number of marginals.
            is_continuous : array_like
                List of boolean values of length d, where d is the number of
                marginals and element i is `True` if marginal i is continuous.
            trunc_level : int, optional
                Layer level to truncate the vine at.  Copulas in layers beyond
                are just independence copulas.  If the level is `None`, then
                the vine is not truncated.  (Default: `None`)

            Returns
            -------
            output_urvs : array_like
                The output uniform random variates of the layer.  Can be
                ignored if this is the output layer.
            '''
            if self.is_marginal_layer():
                output_urvs = np.zeros(samples.shape)
                for i in range(samples.shape[1]):
                    self.marginals[i] = Marginal.fit(samples[:, i],
                                                     is_continuous[i])
                    output_urvs[:, i] = self.marginals[i].cdf(samples[:, i])
            else:
                input_urvs = self.input_layer.fit(samples, is_continuous,
                                                  trunc_level)
                truncate = trunc_level and samples.shape[1] \
                    - len(self.input_indices) > trunc_level - 1
                output_urvs = np.zeros((samples.shape[0],
                                        len(self.input_indices)))
                for i, i_ind in enumerate(self.input_indices):
                    if truncate:
                        self.copulas[i] = IndependenceCopula()
                    else:
                        self.copulas[i] = Copula.fit(input_urvs[:, i_ind])
                    output_urvs[:, i] \
                        = self.copulas[i].ccdf(input_urvs[:, i_ind])
            return output_urvs

        def get_all_params(self):
            '''
            Constructs an array containing all copula parameters.

            Returns
            -------
            params : list
                A list containing all copula parameter values starting with the
                parameters of the first copula layer and continuing layer by
                layer.
            '''
            if self.is_marginal_layer():
                params = []
            else:
                params = self.input_layer.get_all_params()
                for copula in self.copulas:
                    if copula.theta is not None:
                        for param in copula.theta:
                            params.append(param)
            return params

        def set_all_params(self, params):
            '''
            Sets all copula parameters to the values stored in params.

            Parameters
            ----------
            params : list
                A list containing all copula parameter values starting with the
                parameters of the first copula layer and continuing layer by
                layer.
            '''
            if not self.is_marginal_layer():
                self.input_layer.set_all_params(params)
                for i in range(len(self.copulas)):
                    if self.copulas[i].theta is not None:
                        for j in range(len(self.copulas[i].theta)):
                            self.copulas[i].theta[j] = params.pop(0)

        def get_all_bounds(self):
            '''
            Collects the bounds of all copula parameters.

            Returns
            -------
            bnds : list
                A list of 2-tuples containing all copula parameter bounds
                starting with the parameters of the first copula layer and
                continuing layer by layer.  The first element of tuple i
                denotes the lower bound and the second element denotes the
                upper bound of parameter i.
            '''
            if self.is_marginal_layer():
                bnds = []
            else:
                bnds = self.input_layer.get_all_bounds()
                for copula in self.copulas:
                    for bnd in copula.theta_bounds():
                        bnds.append(bnd)
            return bnds

        def is_continuous(self):
            '''
            Determines which marginals are continuous.

            Returns
            -------
            vals : array_like
                List of boolean values of length d, where d is the number of
                marginals and element i is `True` if marginal i is continuous.
            '''
            if self.is_marginal_layer():
                vals = np.zeros(len(self.marginals), dtype=bool)
                for k, marginal in enumerate(self.marginals):
                    vals[k] = marginal.is_continuous
                return vals
            else:
                return self.input_layer.is_continuous()

    def __init__(self, dim):
        if dim < 2:
            raise ValueError("the number of marginals 'dim' must be greater"
                             " than 1")
        self.root = self._construct_c_vine(np.arange(dim))

    def logpdf(self, samples):
        '''
        Calculates the log of the probability density function.

        Parameters
        ----------
        samples : array_like
            n-by-d matrix of samples where n is the number of samples and d is
            the number of marginals.

        Returns
        -------
        ndarray
            Log of the probability density function evaluated at `samples`.
        '''
        return self.root.logpdf(samples)

    def pdf(self, samples):
        '''
        Calculates the probability density function.

        Parameters
        ----------
        samples : array_like
            n-by-d matrix of samples where n is the number of samples and d
            is the number of marginals.

        Returns
        -------
        ndarray
            Probability density function evaluated at `samples`.
        '''
        return np.exp(self.logpdf(samples))

    def rvs(self, size=1, random_state=None):
        '''
        Generates random variates from the mixed vine.

        Parameters
        ----------
        size : int, optional
            The number of samples to generate.  (Default: 1)
        random_state : {None, int, RandomState, Generator}, optional
            The random state to use for random variate generation.  `None`
            corresponds to the `RandomState` singleton.  For an int, a new
            `RandomState` is generated and seeded.  For a `RandomState` or
            `Generator`, the object is used.  (Default: `None`)

        Returns
        -------
        array_like
            n-by-d matrix of samples where n is the number of samples and d is
            the number of marginals.
        '''
        return self.root.rvs(size=size, random_state=random_state)

    def entropy(self, alpha=0.05, sem_tol=1e-3, mc_size=1000,
                random_state=None):
        '''
        Estimates the entropy of the mixed vine.

        Parameters
        ----------
        alpha : float, optional
            Significance level of the entropy estimate.  (Default: 0.05)
        sem_tol : float, optional
            Maximum standard error as a stopping criterion.  (Default: 1e-3)
        mc_size : int, optional
            Number of samples that are drawn in each iteration of the Monte
            Carlo estimation.  (Default: 1000)
        random_state : {None, int, RandomState, Generator}, optional
            The random state to use for random variate generation.  `None`
            corresponds to the `RandomState` singleton.  For an int, a new
            `RandomState` is genered and seeded.  For a `RandomState` or
            `Generator`, the object is used.  (Default: `None`)

        Returns
        -------
        ent : float
            Estimate of the mixed vine entropy in bits.
        sem : float
            Standard error of the mixed vine entropy estimate in bits.
        '''
        # Gaussian confidence interval for sem_tol and level alpha
        conf = norm.ppf(1 - alpha)
        sem = np.inf
        ent = 0.0
        var_sum = 0.0
        k = 0
        while sem >= sem_tol:
            # Generate samples
            samples = self.rvs(size=mc_size, random_state=random_state)
            logp = self.logpdf(samples)
            log2p = logp[np.isfinite(logp)] / np.log(2)
            k += 1
            # Monte-Carlo estimate of entropy
            ent += (-np.mean(log2p) - ent) / k
            # Estimate standard error
            var_sum += np.sum((-log2p - ent) ** 2)
            sem = conf * np.sqrt(var_sum / (k * mc_size * (k * mc_size - 1)))
        return ent, sem

    def set_marginal(self, marginal_index, rv_mixed):
        '''
        Sets a particular marginal distribution in the mixed vine tree for
        manual construction of a mixed vine model.

        Parameters
        ----------
        marginal_index : int
            The index of the marginal in the marginal layer.
        rv_mixed : scipy.stats.distributions.rv_frozen
            The marginal distribution to be inserted.
        '''
        layer = self.root
        while not layer.is_marginal_layer():
            layer = layer.input_layer
        layer.marginals[marginal_index] = Marginal(rv_mixed)

    def set_copula(self, layer_index, copula_index, copula):
        '''
        Sets a particular pair copula in the mixed vine tree for manual
        construction of a mixed vine model.

        Parameters
        ----------
        layer_index : int
            The index of the vine layer.
        copula_index : int
            The index of the copula in its layer.
        copula : Copula
            The copula to be inserted.
        '''
        layer = self.root
        while not layer.is_marginal_layer():
            layer = layer.input_layer
        if layer_index < 1 or layer_index >= len(layer.marginals):
            raise IndexError("argument 'layer_index' out of range")
        for _ in range(layer_index):
            layer = layer.output_layer
        layer.copulas[copula_index] = copula

    def is_continuous(self):
        '''
        Determines which marginals are continuous.

        Returns
        -------
        is_continuous : array_like
            List of boolean values of length d, where d is the number of
            marginals and element i is `True` if marginal i is continuous.
        '''
        return self.root.is_continuous()

    @staticmethod
    def fit(samples, is_continuous, trunc_level=None, do_refine=False,
            keep_order=False):
        '''
        Fits the mixed vine to the given samples.

        Parameters
        ----------
        samples : array_like
            n-by-d matrix of samples where n is the number of samples and d is
            the number of marginals.
        is_continuous : array_like
            List of boolean values of length d, where d is the number of
            marginals and element i is `True` if marginal i is continuous.
        trunc_level : int, optional
            Layer level to truncate the vine at.  Copulas in layers beyond are
            just independence copulas.  If the level is `None`, then the vine
            is not truncated.  (Default: `None`)
        do_refine : boolean, optional
            If `True`, then all pair copula parameters are optimized jointly at
            the end.  (Default: `False`)
        keep_order : boolean, optional
            If `False`, then a heuristic is used to select the vine structure.
            (Default: `False`)

        Returns
        -------
        vine : MixedVine
            The mixed vine with parameters fitted to `samples`.
        '''
        dim = samples.shape[1]
        vine = MixedVine(dim)
        if not keep_order:
            element_order = MixedVine._heuristic_element_order(samples)
            vine.root = MixedVine._construct_c_vine(element_order)
        vine.root.fit(samples, is_continuous, trunc_level)
        if do_refine:
            # Refine copula parameters
            initial_point = vine.root.get_all_params()
            bnds = vine.root.get_all_bounds()

            def cost(params):
                '''
                Calculates the cost of a given set of copula parameters.
                '''
                vine.root.set_all_params(params.tolist())
                vals = vine.logpdf(samples)
                return -np.sum(vals)

            result = minimize(cost, initial_point, method='SLSQP',
                              bounds=bnds)
            vine.root.set_all_params(result.x.tolist())
        return vine

    @staticmethod
    def _heuristic_element_order(samples):
        '''
        Finds an order of elements that heuristically facilitates vine
        modelling.  For this purpose, Kendall's tau is calculated between
        samples of pairs of elements and elements are scored according to the
        sum of absolute Kendall's taus of pairs the elements appear in.

        Parameters
        ----------
        samples : array_like
            n-by-d matrix of samples where n is the number of samples and d is
            the number of marginals.

        Returns
        -------
        order : array_like
            Permutation of all element indices reflecting descending scores.
        '''
        dim = samples.shape[1]
        # Score elements according to total absolute Kendall's tau
        score = np.zeros(dim)
        for i in range(1, dim):
            for j in range(i):
                tau, _ = kendalltau(samples[:, i], samples[:, j])
                score[i] += np.abs(tau)
                score[j] += np.abs(tau)
        # Get order indices for descending score
        order = score.argsort()[::-1]
        return order

    @staticmethod
    def _construct_c_vine(element_order):
        '''
        Constructs a c-vine tree without setting marginals or copulas.  The
        c-vine tree is constructed according to the input element order.  The
        index of the element with the most important dependencies should come
        first in the input argument.

        Parameters
        ----------
        element_order : array_like
            Permutation of all element indices.

        Returns
        -------
        root : VineLayer
            The root layer of the canonical vine tree.
        '''
        dim = len(element_order)
        marginals = np.empty(dim, dtype=Marginal)
        layer = MixedVine.VineLayer(marginals=marginals)
        identity_order = np.arange(dim - 1)
        for i in range(1, dim):
            input_indices = []
            if i == 1:
                order = element_order
            else:
                order = identity_order
            # For each successor layer, generate c-vine input indices
            for j in range(dim - i):
                input_indices.append(np.array([order[0], order[j+1]]))
            copulas = np.empty(len(input_indices), dtype=Copula)
            # Generate vine layer
            layer = MixedVine.VineLayer(input_layer=layer,
                                        input_indices=input_indices,
                                        copulas=copulas)
        root = layer
        return root



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2019, 2021 Arno Onken
#
# This file is part of the mixedvines package.
#
# The mixedvines package is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# The mixedvines package is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
'''
This module implements copula distributions.
'''
import sys
import abc
from scipy.optimize import minimize
from scipy.stats import norm, uniform, multivariate_normal
import numpy as np


class Copula(abc.ABC):
    '''
    This abstract class represents a copula.

    Parameters
    ----------
    theta : array_like, optional
        Parameter array of the copula.  The number of elements depends on the
        copula family.  (Default: `None`)
    rotation : string, optional
        Clockwise rotation of the copula.  Can be one of the elements of
        `Copula.rotation_options` or `None`.  (Default: `None`)

    Attributes
    ----------
    theta : array_like
        Parameter array of the copula.
    rotation : string
        Clockwise rotation of the copula.

    Methods
    -------
    logpdf(samples)
        Log of the probability density function.
    pdf(samples)
        Probability density function.
    logcdf(samples)
        Log of the cumulative distribution function.
    cdf(samples)
        Cumulative distribution function.
    ccdf(samples, axis)
        Conditional cumulative distribution function.
    ppcf(samples, axis)
        Inverse of the conditional cumulative distribution function.
    rvs(size=1, random_state)
        Generate random variates.
    estimate_theta(samples)
        Estimates the `theta` parameters from the given samples.
    fit(samples)
        Fit a copula to samples.
    theta_bounds()
        Bounds for `theta` parameters.
    '''
    rotation_options = ['90°', '180°', '270°']

    def __init__(self, theta=None, rotation=None):
        self.__check_theta(theta)
        self.__check_rotation(rotation)
        self.theta = theta
        self.rotation = rotation

    @classmethod
    def __check_theta(cls, theta):
        '''
        Checks the `theta` parameter.

        Parameters
        ----------
        theta : array_like
            Parameter array of the copula.  The number of elements depends on
            the copula family.
        '''
        bnds = cls.theta_bounds()
        if len(bnds) > 0:
            theta = np.asarray(theta)
            if theta.size != len(bnds):
                raise ValueError("the number of elements of 'theta' does not"
                                 " match the number of family parameters")
            if theta.size == 1:
                if theta < bnds[0][0] or theta > bnds[0][1]:
                    raise ValueError("parameter theta out of bounds")
            else:
                for i, bnd in enumerate(bnds):
                    if theta[i] < bnd[0] or theta[i] > bnd[1]:
                        raise ValueError("parameter theta[" + str(i)
                                         + "] out of bounds")
        elif theta is not None:
            raise ValueError("for this copula family, 'theta' must be 'None'")

    @classmethod
    def __check_rotation(cls, rotation):
        '''
        Checks the `rotation` parameter.

        Parameters
        ----------
        rotation : string
            Rotation of the copula.  Can be one of the elements of
            `Copula.rotation_options` or `None`.
        '''
        if rotation is not None and rotation not in cls.rotation_options:
            raise ValueError("rotation '" + rotation + "' not supported")

    @staticmethod
    def __crop_input(samples):
        '''
        Crops the input to the unit hypercube.  The input is changed and a
        reference to the input is returned.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        samples : array_like
            n-by-2 matrix of cropped samples where n is the number of samples.
        '''
        samples[samples < 0] = 0
        samples[samples > 1] = 1
        return samples

    def __rotate_input(self, samples):
        '''
        Preprocesses the input to account for the copula rotation.  The input
        is changed and a reference to the input is returned.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        samples : array_like
            n-by-2 matrix of rotated samples where n is the number of samples.
        '''
        if self.rotation == '90°':
            samples[:, 1] = 1 - samples[:, 1]
        elif self.rotation == '180°':
            samples[:, :] = 1 - samples[:, :]
        elif self.rotation == '270°':
            samples[:, 0] = 1 - samples[:, 0]
        return samples

    @abc.abstractmethod
    def _logpdf(self, samples):
        '''
        Calculates the log of the probability density function.  The samples
        can be assumed to lie within the unit hypercube.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Log of the probability density function evaluated at `samples`.
        '''
        pass

    def logpdf(self, samples):
        '''
        Calculates the log of the probability density function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Log of the probability density function evaluated at `samples`.
        '''
        samples = np.copy(np.asarray(samples))
        samples = self.__rotate_input(samples)
        inner = np.all(np.bitwise_and(samples > 0.0, samples < 1.0), axis=1)
        outer = np.invert(inner)
        vals = np.zeros(samples.shape[0])
        vals[inner] = self._logpdf(samples[inner, :])
        # Assign zero mass to border
        vals[outer] = -np.inf
        return vals

    def pdf(self, samples):
        '''
        Calculates the probability density function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Probability density function evaluated at `samples`.
        '''
        return np.exp(self.logpdf(samples))

    @abc.abstractmethod
    def _logcdf(self, samples):
        '''
        Calculates the log of the cumulative distribution function.  The
        samples can be assumed to lie within the unit hypercube.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Log of the cumulative distribution function evaluated at `samples`.
        '''
        pass

    def logcdf(self, samples):
        '''
        Calculates the log of the cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Log of the cumulative distribution function evaluated at `samples`.
        '''
        samples = np.copy(np.asarray(samples))
        samples = self.__crop_input(samples)
        samples = self.__rotate_input(samples)
        vals = self._logcdf(samples)
        # Transform according to rotation, but take `__rotate_input` into
        # account.
        if self.rotation == '90°':
            old_settings = np.seterr(divide='ignore')
            vals = np.log(np.maximum(0, samples[:, 0] - np.exp(vals)))
            np.seterr(**old_settings)
        elif self.rotation == '180°':
            old_settings = np.seterr(divide='ignore')
            vals = np.log(np.maximum(0,
                                     (1 - samples[:, 0]) + (1 - samples[:, 1])
                                     - 1.0 + np.exp(vals)))
            np.seterr(**old_settings)
        elif self.rotation == '270°':
            old_settings = np.seterr(divide='ignore')
            vals = np.log(np.maximum(0, samples[:, 1] - np.exp(vals)))
            np.seterr(**old_settings)
        return vals

    def cdf(self, samples):
        '''
        Calculates the cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        ndarray
            Cumulative distribution function evaluated at `samples`.
        '''
        return np.exp(self.logcdf(samples))

    def __axis_wrapper(self, fun, samples, axis):
        '''
        Calls function `fun` with `samples` as argument, but eventually changes
        rotation and samples such that `axis == 0` corresponds to `axis == 1`.

        Parameters
        ----------
        fun : function
            Function to be called with `samples` as argument.
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.
        axis : int
            The axis to condition the cumulative distribution function on.

        Returns
        -------
        vals : array_like
            Function values evaluated at `samples` but taking `axis` into
            account.
        '''
        samples = np.copy(np.asarray(samples))
        samples = self.__crop_input(samples)
        rotation = self.rotation
        try:
            # Temporarily change rotation according to axis
            if axis == 0:
                if self.rotation == '90°':
                    self.rotation = '270°'
                elif self.rotation == '270°':
                    self.rotation = '90°'
                samples = samples[:, [1, 0]]
            elif axis != 1:
                raise ValueError("axis must be in [0, 1]")
            samples = self.__rotate_input(samples)
            vals = fun(samples)
            if self.rotation == '180°' or self.rotation == '270°':
                vals = 1.0 - vals
        finally:
            # Recover original rotation
            self.rotation = rotation
        return vals

    @abc.abstractmethod
    def _ccdf(self, samples):
        '''
        Calculates the conditional cumulative distribution function conditioned
        on axis 1.  The samples can be assumed to lie within the unit
        hypercube.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Conditional cumulative distribution function evaluated at
            `samples`.
        '''
        pass

    def ccdf(self, samples, axis=1):
        '''
        Calculates the conditional cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.
        axis : int, optional
            The axis to condition the cumulative distribution function on.
            (Default: 1)

        Returns
        -------
        vals : ndarray
            Conditional cumulative distribution function evaluated at
            `samples`.
        '''
        return self.__axis_wrapper(self._ccdf, samples, axis)

    @abc.abstractmethod
    def _ppcf(self, samples):
        '''
        Calculates the inverse of the copula conditional cumulative
        distribution function conditioned on axis 1.  The samples can be
        assumed to lie within the unit hypercube.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        vals : ndarray
            Inverse of the conditional cumulative distribution function
            evaluated at `samples`.
        '''
        pass

    def ppcf(self, samples, axis=1):
        '''
        Calculates the inverse of the copula conditional cumulative
        distribution function.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.
        axis : int, optional
            The axis to condition the cumulative distribution function on.
            (Default: 1)

        Returns
        -------
        vals : ndarray
            Inverse of the conditional cumulative distribution function
            evaluated at `samples`.
        '''
        return self.__axis_wrapper(self._ppcf, samples, axis)

    def rvs(self, size=1, random_state=None):
        '''
        Generates random variates from the copula.

        Parameters
        ----------
        size : int, optional
            The number of samples to generate.  (Default: 1)
        random_state : {None, int, RandomState, Generator}, optional
            The random state to use for random variate generation.  `None`
            corresponds to the `RandomState` singleton.  For an int, a new
            `RandomState` is generated and seeded.  For a `RandomState` or
            `Generator`, the object is used.  (Default: `None`)

        Returns
        -------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.
        '''
        samples = np.stack((uniform.rvs(size=size, random_state=random_state),
                            uniform.rvs(size=size, random_state=random_state)),
                           axis=1)
        samples[:, 0] = self.ppcf(samples)
        return samples

    def estimate_theta(self, samples):
        '''
        Estimates the theta parameters from the given samples.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.
        '''
        if self.theta is not None:
            bnds = self.theta_bounds()

            def cost(theta):
                '''
                Calculates the cost of a given `theta` parameter.
                '''
                self.theta = np.asarray(theta)
                vals = self.logpdf(samples)
                # For optimization, filter out inifinity values
                return -np.sum(vals[np.isfinite(vals)])

            result = minimize(cost, self.theta, method='TNC', bounds=bnds)
            self.theta = result.x

    @classmethod
    def fit(cls, samples):
        '''
        Fits the parameters of the copula to the given samples.

        Parameters
        ----------
        samples : array_like
            n-by-2 matrix of samples where n is the number of samples.

        Returns
        -------
        copula : Copula
            The copula fitted to `samples`.
        '''
        # Find best fitting family
        copulas = []
        for family in cls.__subclasses__():
            copulas.append(family.fit(samples))
        # Calculate Akaike information criterion
        aic = np.zeros(len(copulas))
        for i, copula in enumerate(copulas):
            aic[i] = - 2 * np.sum(copula.logpdf(samples))
            if copula.theta is not None:
                if np.isscalar(copula.theta):
                    aic[i] += 2
                else:
                    aic[i] += 2 * len(copula.theta)
        copula = copulas[np.argmin(aic)]
        return copula

    @staticmethod
    @abc.abstractmethod
    def theta_bounds():
        '''
        Bounds for `theta` parameters.

        Returns
        -------
        bnds : array_like
            List of 2-tuples where the first tuple element represents the lower
            bound and the second element represents the upper bound.
        '''
        pass


class IndependenceCopula(Copula):
    '''
    This class represents the independence copula.
    '''

    def _logpdf(self, samples):
        vals = np.zeros(samples.shape[0])
        return vals

    def _logcdf(self, samples):
        old_settings = np.seterr(divide='ignore')
        vals = np.sum(np.log(samples), axis=1)
        np.seterr(**old_settings)
        return vals

    def _ccdf(self, samples):
        vals = samples[:, 0]
        return vals

    def _ppcf(self, samples):
        vals = samples[:, 0]
        return vals

    @classmethod
    def fit(cls, samples):
        copula = cls()
        return copula

    @staticmethod
    def theta_bounds():
        bnds = []
        return bnds


class GaussianCopula(Copula):
    '''
    This class represents a copula from the Gaussian family.
    '''

    def _logpdf(self, samples):
        if self.theta >= 1.0:
            vals = np.zeros(samples.shape[0])
            vals[samples[:, 0] == samples[:, 1]] = np.inf
        elif self.theta <= -1.0:
            vals = np.zeros(samples.shape[0])
            vals[samples[:, 0] == 1 - samples[:, 1]] = np.inf
        else:
            nrvs = norm.ppf(samples)
            vals = 2 * self.theta * nrvs[:, 0] * nrvs[:, 1] - self.theta**2 \
                * (nrvs[:, 0]**2 + nrvs[:, 1]**2)
            vals /= 2 * (1 - self.theta**2)
            vals -= np.log(1 - self.theta**2) / 2
        return vals

    def _logcdf(self, samples):
        upper = norm.ppf(samples)
        cov = [[1.0, self.theta], [self.theta, 1.0]]
        vals = multivariate_normal.logcdf(upper, None, cov)
        vals[np.any(samples == 0.0, axis=1)] = -np.inf
        vals[samples[:, 0] == 1.0] = np.log(samples[samples[:, 0] == 1.0, 1])
        vals[samples[:, 1] == 1.0] = np.log(samples[samples[:, 1] == 1.0, 0])
        return vals

    def _ccdf(self, samples):
        vals = np.zeros(samples.shape[0])
        # Avoid subtraction of infinities
        neqz = np.bitwise_and(np.any(samples > 0.0, axis=1),
                              np.any(samples < 1.0, axis=1))
        nrvs = norm.ppf(samples[neqz, :])
        vals[neqz] = norm.cdf((nrvs[:, 0] - self.theta * nrvs[:, 1])
                              / np.sqrt(1 - self.theta**2))
        vals[np.invert(neqz)] = norm.cdf(0.0)
        return vals

    def _ppcf(self, samples):
        nrvs = norm.ppf(samples)
        vals = norm.cdf(nrvs[:, 0] * np.sqrt(1 - self.theta**2)
                        + self.theta * nrvs[:, 1])
        return vals

    @classmethod
    def fit(cls, samples):
        initial_point = (0.0)
        copula = cls(theta=initial_point)
        copula.estimate_theta(samples)
        return copula

    @staticmethod
    def theta_bounds():
        bnds = [(-1.0 + 1e-3, 1.0 - 1e-3)]
        return bnds


class ClaytonCopula(Copula):
    '''
    This class represents a copula from the Clayton family.
    '''

    def _logpdf(self, samples):
        if self.theta == 0:
            vals = np.zeros(samples.shape[0])
        else:
            vals = np.log(1 + self.theta) + (-1 - self.theta) \
                   * (np.log(samples[:, 0]) + np.log(samples[:, 1])) \
                   + (-1 / self.theta - 2) \
                   * np.log(samples[:, 0]**(-self.theta)
                            + samples[:, 1]**(-self.theta) - 1)
        return vals

    def _logcdf(self, samples):
        if self.theta == 0:
            vals = np.sum(np.log(samples), axis=1)
        else:
            old_settings = np.seterr(divide='ignore')
            vals = (-1 / self.theta) \
                * np.log(np.maximum(samples[:, 0]**(-self.theta)
                                    + samples[:, 1]**(-self.theta) - 1, 0))
            np.seterr(**old_settings)
        return vals

    def _ccdf(self, samples):
        if self.theta == 0:
            vals = samples[:, 0]
        else:
            vals = np.zeros(samples.shape[0])
            gtz = np.all(samples > 0.0, axis=1)
            vals[gtz] = np.maximum(samples[gtz, 1]**(-1 - self.theta)
                                   * (samples[gtz, 0]**(-self.theta)
                                      + samples[gtz, 1]**(-self.theta) - 1)
                                   ** (-1 - 1 / self.theta), 0)
        return vals

    def _ppcf(self, samples):
        if self.theta == 0:
            vals = samples[:, 0]
        else:
            vals = np.zeros(samples.shape[0])
            gtz = np.all(samples > 0.0, axis=1)
            vals[gtz] = (1 - samples[gtz, 1]**(-self.theta)
                         + (samples[gtz, 0]
                            * (samples[gtz, 1]**(1 + self.theta)))
                         ** (-self.theta / (1 + self.theta))) \
                ** (-1 / self.theta)
        return vals

    @classmethod
    def fit(cls, samples):
        initial_point = (1.0)
        # Optimize rotation as well
        copulas = [cls(theta=initial_point)]
        for rotation in cls.rotation_options:
            copulas.append(cls(theta=initial_point, rotation=rotation))
        # Fit parameters and calculate Akaike information criterion
        aic = np.zeros(len(copulas))
        for i, _ in enumerate(copulas):
            copulas[i].estimate_theta(samples)
            aic[i] = - 2 * np.sum(copulas[i].logpdf(samples)) \
                + 2 * len(copulas[i].theta)
        # Select best copula
        copula = copulas[np.argmin(aic)]
        return copula

    @staticmethod
    def theta_bounds():
        bnds = [(1e-3, 20)]
        return bnds


class FrankCopula(Copula):
    '''
    This class represents a copula from the Frank family.
    '''

    def _logpdf(self, samples):
        if self.theta == 0:
            vals = np.zeros(samples.shape[0])
        else:
            vals = np.log(-self.theta * np.expm1(-self.theta)
                          * np.exp(-self.theta
                                   * (samples[:, 0] + samples[:, 1]))
                          / (np.expm1(-self.theta)
                             + np.expm1(-self.theta * samples[:, 0])
                             * np.expm1(-self.theta * samples[:, 1])) ** 2)
        return vals

    def _logcdf(self, samples):
        if self.theta == 0:
            vals = np.sum(np.log(samples), axis=1)
        else:
            old_settings = np.seterr(divide='ignore')
            vals = np.log(-np.log1p(np.expm1(-self.theta * samples[:, 0])
                                    * np.expm1(-self.theta * samples[:, 1])
                                    / (np.expm1(-self.theta)))) \
                - np.log(self.theta)
            np.seterr(**old_settings)
        return vals

    def _ccdf(self, samples):
        if self.theta == 0:
            vals = samples[:, 0]
        else:
            vals = np.exp(-self.theta * samples[:, 1]) \
                * np.expm1(-self.theta * samples[:, 0]) \
                / (np.expm1(-self.theta)
                   + np.expm1(-self.theta * samples[:, 0])
                   * np.expm1(-self.theta * samples[:, 1]))
        return vals

    def _ppcf(self, samples):
        if self.theta == 0:
            vals = samples[:, 0]
        else:
            vals = -np.log1p(samples[:, 0] * np.expm1(-self.theta)
                             / (np.exp(-self.theta * samples[:, 1])
                                - samples[:, 0] * np.expm1(-self.theta
                                                           * samples[:, 1]))) \
                / self.theta
        return vals

    @classmethod
    def fit(cls, samples):
        initial_point = (0.0)
        copula = cls(theta=initial_point)
        copula.estimate_theta(samples)
        return copula

    @staticmethod
    def theta_bounds():
        bnds = [(-20, 20)]
        return bnds



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2019, 2021 Arno Onken
#
# This file is part of the mixedvines package.
#
# The mixedvines package is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# The mixedvines package is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
'''
This module implements univariate marginal distribution that are either
continuous or discrete.
'''
from scipy.stats import rv_continuous, norm, gamma, poisson, binom, nbinom
import numpy as np


class Marginal(object):
    '''
    This class represents a marginal distribution which can be continuous or
    discrete.

    Parameters
    ----------
    rv_mixed : scipy.stats.distributions.rv_frozen
        The distribution object, either of a continuous or of a discrete
        univariate distribution.

    Attributes
    ----------
    rv_mixed : scipy.stats.distributions.rv_frozen
        The distribution object.
    is_continuous : boolean
        `True` if the distribution is continuous.

    Methods
    -------
    logpdf(samples)
        Log of the probability density function or probability mass function.
    pdf(samples)
        Probability density function or probability mass function.
    logcdf(samples)
        Log of the cumulative distribution function.
    cdf(samples)
        Cumulative distribution function.
    ppf(samples)
        Inverse of the cumulative distribution function.
    rvs(size, random_state)
        Generate random variates.
    fit(samples, is_continuous)
        Fit a distribution to samples.
    '''

    def __init__(self, rv_mixed):
        self.rv_mixed = rv_mixed
        self.is_continuous = isinstance(rv_mixed.dist, rv_continuous)

    def logpdf(self, samples):
        '''
        Calculates the log of the probability density function.

        Parameters
        ----------
        samples : array_like
            Array of samples.

        Returns
        -------
        ndarray
            Log of the probability density function evaluated at `samples`.
        '''
        if self.is_continuous:
            return self.rv_mixed.logpdf(samples)
        return self.rv_mixed.logpmf(samples)

    def pdf(self, samples):
        '''
        Calculates the probability density function.

        Parameters
        ----------
        samples : array_like
            Array of samples.

        Returns
        -------
        ndarray
            Probability density function evaluated at `samples`.
        '''
        return np.exp(self.logpdf(samples))

    def logcdf(self, samples):
        '''
        Calculates the log of the cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            Array of samples.

        Returns
        -------
        ndarray
            Log of the cumulative distribution function evaluated at `samples`.
        '''
        return self.rv_mixed.logcdf(samples)

    def cdf(self, samples):
        '''
        Calculates the cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            Array of samples.

        Returns
        -------
        ndarray
            Cumulative distribution function evaluated at `samples`.
        '''
        return np.exp(self.logcdf(samples))

    def ppf(self, samples):
        '''
        Calculates the inverse of the cumulative distribution function.

        Parameters
        ----------
        samples : array_like
            Array of samples.

        Returns
        -------
        ndarray
            Inverse of the cumulative distribution function evaluated at
            `samples`.
        '''
        return self.rv_mixed.ppf(samples)

    def rvs(self, size=1, random_state=None):
        '''
        Generates random variates from the distribution.

        Parameters
        ----------
        size : int, optional
            The number of samples to generate.  (Default: 1)
        random_state : {None, int, RandomState, Generator}, optional
            The random state to use for random variate generation.  `None`
            corresponds to the `RandomState` singleton.  For an int, a new
            `RandomState` is generated and seeded.  For a `RandomState` or
            `Generator`, the object is used.  (Default: `None`)

        Returns
        -------
        array_like
            Array of samples.
        '''
        return self.rv_mixed.rvs(size, random_state=random_state)

    @staticmethod
    def fit(samples, is_continuous):
        '''
        Fits a distribution to the given samples.

        Parameters
        ----------
        samples : array_like
            Array of samples.
        is_continuous : bool
            If `True` then a continuous distribution is fitted.  Otherwise, a
            discrete distribution is fitted.

        Returns
        -------
        best_marginal : Marginal
            The distribution fitted to `samples`.
        '''
        # Mean and variance
        mean = np.mean(samples)
        var = np.var(samples)
        # Set suitable distributions
        if is_continuous:
            if np.any(samples <= 0):
                options = [norm]
            else:
                options = [norm, gamma]
        else:
            if var > mean:
                options = [poisson, binom, nbinom]
            else:
                options = [poisson, binom]
        params = np.empty(len(options), dtype=object)
        marginals = np.empty(len(options), dtype=object)
        # Fit parameters and construct marginals
        for i, dist in enumerate(options):
            if dist == poisson:
                params[i] = [mean]
            elif dist == binom:
                param_n = np.max(samples)
                param_p = np.sum(samples) / (param_n * len(samples))
                params[i] = [param_n, param_p]
            elif dist == nbinom:
                param_n = mean * mean / (var - mean)
                param_p = mean / var
                params[i] = [param_n, param_p]
            else:
                params[i] = dist.fit(samples)
            rv_mixed = dist(*params[i])
            marginals[i] = Marginal(rv_mixed)
        # Calculate Akaike information criterion
        aic = np.zeros(len(options))
        for i, marginal in enumerate(marginals):
            aic[i] = 2 * len(params[i]) \
                     - 2 * np.sum(marginal.logpdf(samples))
        best_marginal = marginals[np.argmin(aic)]
        return best_marginal



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2019, 2021 Arno Onken
#
# This file is part of the mixedvines package.
#
# The mixedvines package is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# The mixedvines package is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Setup for the mixedvines package.
'''
import setuptools

setuptools.setup()
